/*
 * Copyright (c) 1998-2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 *
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * Copyright (c) 2000-2001 Apple Computer, Inc.  All rights reserved.
 *
 *  DRI: Josh de Cesare
 *
 */

#include <libkern/OSByteOrder.h>
#include <IOKit/IOLib.h>
#include <IOKit/graphics/IOFramebuffer.h>

#include "AppleBt8xx.h"
#include "AppleBt8xxRegs.h"


//---------------------------------------------------------------------------//

void AppleBt8xx::_identifyChipType(IOPCIDevice *pciNub)
{
  UInt16 deviceID = pciNub->configRead16(kIOPCIConfigDeviceID);
  UInt8  revision = pciNub->configRead8(kIOPCIConfigRevisionID);
  
  switch (deviceID) {
    case kBtDeviceID848VF :
      if (revision < kBt848ARevision) chipType = kBt848;
      else chipType = kBt848A;
      break;
    
    case kBtDeviceID849VF  : chipType = kBt849A; break;
    case kBtDeviceID878VF  : chipType = kBt878; break;
    case kBtDeviceID878aVF : chipType = kBt878; break;
    case kBtDeviceID878AF  : chipType = kBt878; break;
    case kBtDeviceID879VF  : chipType = kBt879; break;
    case kBtDeviceID879AF  : chipType = kBt879; break;
    default 	      	   : chipType = kBt8xx; break;
    }
}

IOReturn AppleBt8xx::_processCmd(AppleBt8xxCmdPtr /*cmd*/)
{
#if 0  
  IOReturn ret;
  UInt32   tmp;

  switch (cmd->type) {
  case kAppleBt8xxCmdNone :
    ret = kIOReturnSuccess;
    break;
    
  case kAppleBt8xxCmdSetup :
    ret = _setup();
    break;
    
  case kAppleBt8xxCmdStart :
    ret = _start();
    break;
    
  case kAppleBt8xxCmdStop :
    ret = _stop();
    break;
    
  case kAppleBt8xxCmdSetSource :
    ret = _setSource(cmd->data1);
    break;
    
  case kAppleBt8xxCmdTune :
    ret = _setTuner(cmd->data1);
    break;
    
  case kAppleBt8xxCmdReadReg :
    if (cmd->data1 == 1) tmp = (UInt32)_readVideoReg8((UInt32)cmd->data2);
    else tmp = _readVideoReg32((UInt32)cmd->data2);
    setProperty("Register-Data", tmp, 32);
    ret = kIOReturnSuccess;
    break;
    
  case kAppleBt8xxCmdWriteReg :
    if (cmd->data1 == 1) _writeVideoReg8((UInt32)cmd->data2,(UInt8)cmd->data3);
    else _writeVideoReg32((UInt32)cmd->data2, (UInt32)cmd->data3);
    ret = kIOReturnSuccess;
    break;
    
  default :
    ret = kIOReturnUnsupported;
    break;
  }
#endif
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_reset(void)
{
  _currentMode   = kBtModeReset;
  _currentSource = kBtSourceNone;
  _currentFormat = kBtFormatNone;
  
  _currentSource = -1;
  _currentSourceProperties = 0;
  _currentSourceHasVideo = false;
  _currentSourceHasAudio = false;
  
  _writeVideoReg8(kBtSReset, 0);
  
  _writeVideoReg8(kBtADC, kBtADCReserved | kBtADCCrush);
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_setup(void)
{
  OSObject           *tmpObject;
  OSIterator         *frameBuffers;
  IOFramebuffer      *frameBuffer;
  IODeviceMemory     *frameBufferRange;
  IOPhysicalAddress  frameBufferAddress;
  Bounds             *bounds;
  IODisplayModeID    mode;
  IOIndex            depth;
  IOPixelInformation info;
  UInt8              pixelMode, colorCtl;
  
  if (_currentMode == kBtModeRunning) _stop();
  
  frameBuffers = getMatchingServices(serviceMatching("IOFramebuffer"));
  if (frameBuffers == 0) return kIOReturnError;
  while (tmpObject = frameBuffers->getNextObject()) {
    frameBuffer = OSDynamicCast(IOFramebuffer, tmpObject);
    if (frameBuffer != 0) {
      frameBuffer->getBoundingRect(&bounds);
      if ((bounds->minx == 0) && (bounds->miny == 0)) break;
      frameBuffer = 0;
    }
  }
  frameBuffers->release();
  if (frameBuffer == 0) return kIOReturnError;
  
  frameBuffer->getCurrentDisplayMode(&mode, &depth);
  frameBuffer->getPixelInformation(mode, depth, kIOFBSystemAperture, &info);
  
  frameBufferRange = frameBuffer->getApertureRange(kIOFBSystemAperture);
  frameBufferAddress = frameBufferRange->getPhysicalAddress();
  frameBufferRange->release();
  
  switch (info.bitsPerPixel) {
  case 8  :
    pixelMode = kBtColorFmtRGB8;
    colorCtl = 0;
    break;
    
  case 16 :
    pixelMode = kBtColorFmtRGB15;
    colorCtl = kBtColorCtlBSwapOdd | kBtColorCtlBSwapEven;
    break;
    
  case 32 :
    pixelMode = kBtColorFmtRGB32;
    colorCtl = kBtColorCtlWSwapOdd | kBtColorCtlWSwapEven |
      kBtColorCtlBSwapOdd | kBtColorCtlBSwapEven;
    break;
    
  default :
    return kIOReturnError;
  }
  
  _screenInfo.baseAddress = frameBufferAddress;
  _screenInfo.rowBytes = info.bytesPerRow;
  _screenInfo.pixelSize = (info.bitsPerPixel + 7) / 8;
  _screenInfo.width = info.activeWidth;
  _screenInfo.height = info.activeHeight;
  
  _left   = 0;
  _top    = 0;
  _width  = 640;
  _heigth = 480;
  
  _setSource(0);
  
  _writeVideoReg8(kBtColorFmt,
		  pixelMode << kBtColorFmtEvenShift |
		  pixelMode << kBtColorFmtOddShift);
  
  colorCtl |= kBtColorCtlGamma;
  _writeVideoReg8(kBtColorCtl, colorCtl);
  
  _setupVideoDMA();
  
  _currentMode = kBtModeStopped;
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_start(void)
{
  UInt32 data;
  
  if (_currentMode == kBtModeRunning) return kIOReturnSuccess;
  if (_currentMode == kBtModeReset) _setup();  
  if (_currentMode == kBtModeStopped) {
    // Should probably wait for video present.
    
    if (_currentSourceHasVideo) {
      // Start Video DMA.
      _writeVideoReg32(kBtStartAddr, _videoCmdPagePhys);
      _writeVideoReg32(kBtIntStat, 0x00FFFFFF);
      _writeVideoReg32(kBtGPIODMACTL,
		       0xAC | kBtGPIODMACTLRISCEn | kBtGPIODMACTLFIFOEn);
    }
    
    // Enable Audio Mux.
    if (_currentSourceHasAudio) {
      data = _readVideoReg32(kBtGPIOOutEn);
      data |= _audioMask;
      _writeVideoReg32(kBtGPIOOutEn, data);
    }
    
    _currentMode = kBtModeRunning;
    
    return kIOReturnSuccess;
  }
  
  return kIOReturnError;
}

IOReturn AppleBt8xx::_stop(void)
{
  UInt32 data;
  
  if (_currentMode == kBtModeStopped) return kIOReturnSuccess;
  if (_currentMode == kBtModeRunning) {
    // Stop Video DMA.
    _writeVideoReg32(kBtGPIODMACTL, 0);
    
    // Disable Audio Mux.
    data = _readVideoReg32(kBtGPIOOutEn);
    data &= ~_audioMask;
    _writeVideoReg32(kBtGPIOOutEn, data);
    
    _currentMode = kBtModeStopped;
    
    return kIOReturnSuccess;
  }
  
  return kIOReturnError;
}

OSDictionary *AppleBt8xx::_getSourceProperties(UInt32 source)
{
  UInt32       count;
  OSDictionary *dict = 0;
  
  count = _sources->getCount();
  if (source < count) {
    dict = OSDynamicCast(OSDictionary, _sources->getObject(source));
  }
  
  return dict;
}

IOReturn AppleBt8xx::_setSource(SInt32 source)
{
  UInt32       iform, mux, controlE, controlO, gpio;
  SInt32       videoSource, audioSource;
  Boolean      hasVideo, hasAudio, svideo;
  OSData       *tmpData;
  OSString     *tmpString;
  OSDictionary *newSourceProperties;
  
  if (_currentSource == source) return kIOReturnSuccess;
  newSourceProperties = _getSourceProperties(source);
  if (newSourceProperties == 0) return kIOReturnError; 
  
  // Get the Video Mux source for Video.
  hasVideo = false;
  tmpData = OSDynamicCast(OSData,
			  newSourceProperties->getObject("VideoSource"));
  if (tmpData != 0) {
    videoSource = *(UInt8 *)tmpData->getBytesNoCopy();
    
    hasVideo = true;
    switch (videoSource) {
    case 0  : mux = kBtIFormMuxSelSelMux0; break;
    case 1  : mux = kBtIFormMuxSelSelMux1; break;
    case 2  : mux = kBtIFormMuxSelSelMux2; break;
    case 3  : mux = kBtIFormMuxSelSelMux3; break;
    default : hasVideo = false; break;
    }
    
    // Get the type of this source composite or s-video.
    svideo = false;
    tmpString = OSDynamicCast(OSString, newSourceProperties->getObject("VideoType"));
    if ((tmpString != 0) && tmpString->isEqualTo(kBtSourceTypeSVideo)) {
      svideo = true;
    }
  }
  
  // Get the Audio Mux source for Video.
  hasAudio = false;
  tmpData = OSDynamicCast(OSData,
			  newSourceProperties->getObject("AudioSource"));
  if (tmpData != 0) {
    audioSource = *(UInt32 *)tmpData->getBytesNoCopy();
    hasAudio = true;
  }
  
  _currentSourceProperties = newSourceProperties;
  _currentSource = source;
  
  _currentSourceHasVideo = hasVideo;
  if (hasVideo) {
    iform = _readVideoReg8(kBtIForm);
    iform &= ~kBtIFormMuxSelMask;
    _writeVideoReg8(kBtIForm, iform);
    IOSleep(10);
    
    controlE = _readVideoReg8(kBtControl + kBtFieldEven);
    controlO = _readVideoReg8(kBtControl + kBtFieldOdd);
    if (svideo) {
      controlE |= kBtControlComp;
      controlO |= kBtControlComp;
    } else {
      controlE &= ~kBtControlComp;
      controlO &= ~kBtControlComp;
    }
    _writeVideoReg8(kBtControl + kBtFieldEven, controlE);
    _writeVideoReg8(kBtControl + kBtFieldOdd, controlO);
    
    iform |= mux;
    _writeVideoReg8(kBtIForm, iform);
  }
  
  _currentSourceHasAudio = hasAudio;
  if (hasAudio) {
    // Select the Audio Mux source.
    gpio = _readVideoReg32(kBtGPIOData);
    gpio &= ~_audioMask;
    gpio |= audioSource & _audioMask;
    _writeVideoReg32(kBtGPIOData, gpio);
    
    // Enable the Audio Mux is running.
    if (_currentMode == kBtModeRunning) {
      gpio = _readVideoReg32(kBtGPIOOutEn);
      gpio |= _audioMask;
      _writeVideoReg32(kBtGPIOOutEn, gpio);
    }
  } else {
    // Disable the Audio Mux.
    gpio = _readVideoReg32(kBtGPIOOutEn);
    gpio &= ~_audioMask;
    _writeVideoReg32(kBtGPIOOutEn, gpio);
  }
  
  return kIOReturnSuccess;
}

void AppleBt8xx::_selectFormat(eBtFormat format)
{
  UInt8   data;
  
  if (_currentFormat == format) return;
  _currentFormat = format;
  
  data = _readVideoReg8(kBtIForm);
  data &= ~kBtIFormFormatMask;
  
  switch (format) {
  case kBtFormatAuto   : data |= kBtIFormFormatAuto;   break;
  case kBtFormatNTSC_M : data |= kBtIFormFormatNTSC_M; break;
  case kBtFormatNTSC_J : data |= kBtIFormFormatNTSC_J; break;
  case kBtFormatPAL    : data |= kBtIFormFormatPAL;    break;
  case kBtFormatPAL_M  : data |= kBtIFormFormatPAL_M;  break;
  case kBtFormatPAL_N  : data |= kBtIFormFormatPAL_N;  break;
  case kBtFormatPAL_Nc : data |= kBtIFormFormatPAL_Nc; break;
  case kBtFormatSECAM  : data |= kBtIFormFormatSECAM;  break;
  default : return;
  }
  
  _writeVideoReg8(kBtIForm, data);
}

IOReturn AppleBt8xx::_setTuner(UInt32 freq)
{
  bool     radio;
  UInt8    addr, band, data1, data2;
  UInt32   cnt, data, bands, bandMin, bandMax, intermediateFreq, oscFreq;
  IOReturn ret;
  OSData   *tmpData;
  OSArray  *tmpArray;
  OSString *tmpString;
  
  if (_currentSourceProperties == 0) return kIOReturnError;
  
  tmpData = OSDynamicCast(OSData, _currentSourceProperties->getObject(kBtSourceTunerKey));
  if (tmpData == 0) return kIOReturnError;
  addr = *(UInt8 *)tmpData->getBytesNoCopy();
  
  tmpData = OSDynamicCast(OSData, _currentSourceProperties->getObject("IntermediateFrequency"));
  if (tmpData == 0) return kIOReturnError;
  intermediateFreq = *(UInt32 *)tmpData->getBytesNoCopy();
  
  radio = false;
  tmpString = OSDynamicCast(OSString, _currentSourceProperties->getObject("TunerType"));
  if ((tmpString != 0) && tmpString->isEqualTo("Radio")) {
    radio = true;
  }
  
  tmpArray = OSDynamicCast(OSArray, _currentSourceProperties->getObject("Bands"));
  if (tmpArray == 0) return kIOReturnError;
  band = 0;
  bands = tmpArray->getCount();
  for (cnt = 0; cnt < bands; cnt++) {
    tmpData = OSDynamicCast(OSData, tmpArray->getObject(cnt));
    if ((tmpData == 0) || (tmpData->getLength() != 12))
      return kIOReturnError;
    
    bandMin = ((UInt32 *)tmpData->getBytesNoCopy())[0];
    bandMax = ((UInt32 *)tmpData->getBytesNoCopy())[1];
    if ((bandMin <= freq) && (freq <= bandMax)) {
      band = ((UInt32 *)tmpData->getBytesNoCopy())[2];
    }
  }
  if (band == 0) return kIOReturnError;
  
  oscFreq = ((freq + intermediateFreq) * 4) / 25;

  // Disable the Audio Mux if running.
  if (_currentSourceHasAudio && (_currentMode == kBtModeRunning)) {
    data = _readVideoReg32(kBtGPIOOutEn);
    data &= ~_audioMask;
    _writeVideoReg32(kBtGPIOOutEn, data);
  }
  
  data1 = (oscFreq >> 8) & 0x07F;
  data2 = oscFreq & 0x0FF;  
  ret = _sendIICWrite(addr, data1, data2, true);
  if (ret != kIOReturnSuccess) return ret;
  
  if (radio) data1 = 0xCE;
  else  data1 = 0xCE;
  
  ret = _sendIICWrite(addr, data1, band, true);
  if (ret != kIOReturnSuccess) return ret;
  
  // Enable the Audio Mux if running.
  if (_currentSourceHasAudio && (_currentMode == kBtModeRunning)) {
    data = _readVideoReg32(kBtGPIOOutEn);
    data |= _audioMask;
    _writeVideoReg32(kBtGPIOOutEn, data);
  }
  
  return kIOReturnSuccess;
}


IOReturn AppleBt8xx::_setBrightness(SInt8 brightness)
{
  _writeVideoReg8(kBtBright, brightness);
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_setContrast(UInt16 contrast)
{
  UInt8 dataE, dataO;
  
  _writeVideoReg8(kBtContrastLo, contrast & 0xFF);
  
  dataE = _readVideoReg8(kBtControl + kBtFieldEven);
  dataO = _readVideoReg8(kBtControl + kBtFieldOdd);
  if (contrast & 0x100) {
    dataE |= kBtControlConMSB;
    dataO |= kBtControlConMSB;
  } else {
    dataE &= ~kBtControlConMSB;
    dataO &= ~kBtControlConMSB;
  }
  _writeVideoReg8(kBtControl + kBtFieldEven, dataE);
  _writeVideoReg8(kBtControl + kBtFieldOdd, dataO);
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_setSaturationU(UInt16 saturationU)
{
  UInt8 dataE, dataO;

  _writeVideoReg8(kBtSatULo, saturationU & 0xFF);

  dataE = _readVideoReg8(kBtControl + kBtFieldEven);
  dataO = _readVideoReg8(kBtControl + kBtFieldOdd);
  if (saturationU & 0x100) {
    dataE |= kBtControlSatUMSB;
    dataO |= kBtControlSatUMSB;
  } else {
    dataE &= ~kBtControlSatUMSB;
    dataO &= ~kBtControlSatUMSB;
  }
  _writeVideoReg8(kBtControl + kBtFieldEven, dataE);
  _writeVideoReg8(kBtControl + kBtFieldOdd, dataO);

  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_setSaturationV(UInt16 saturationV)
{
  UInt8 dataE, dataO;

  _writeVideoReg8(kBtSatVLo, saturationV & 0xFF);
  
  dataE = _readVideoReg8(kBtControl + kBtFieldEven);
  dataO = _readVideoReg8(kBtControl + kBtFieldOdd);
  if (saturationV & 0x100) {
    dataE |= kBtControlSatVMSB;
    dataO |= kBtControlSatVMSB;
  } else {
    dataE &= ~kBtControlSatVMSB;
    dataO &= ~kBtControlSatVMSB;
  }
  _writeVideoReg8(kBtControl + kBtFieldEven, dataE);
  _writeVideoReg8(kBtControl + kBtFieldOdd, dataO);

  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_setHue(SInt8 hue)
{
  _writeVideoReg8(kBtHue, hue);
  
  return kIOReturnSuccess;
}


IOReturn AppleBt8xx::_sendIICWrite(UInt8 addr, UInt8 data1,
				     UInt8 data2, bool three)
{
  UInt32   regData, regStatus;
  
  regData = kBtIICModeHardware | kBtIICRate100kHz | kBtIICSCL | kBtIICSDA;
  
  regData |= addr << kBtIICDB0Shift;
  regData |= data1 << kBtIICDB1Shift;
  if (three) {
    regData |= data2 << kBtIICDB2Shift;
    regData |= kBtIICW3BRA;
  }
  
  // Clear the IIC Done bit.
  _writeVideoReg32(kBtIntStat, kBtIntIICDone);
  _writeVideoReg32(kBtIIC, regData);
  
  do {
    regStatus = _readVideoReg32(kBtIntStat);
  } while (!(regStatus & kBtIntIICDone));
  
  if (!(regStatus & kBtStatusRACK)) return kIOReturnError;
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xx::_sendIICRead(UInt8 addr, UInt8 subaddr,
				    UInt8 *data, UInt8 length)
{
  UInt8    cnt, timeOut;
  UInt32   regData, regStatus;
  IOReturn ret = kIOReturnSuccess;
  
  addr |= 1;
  
  for (cnt = 0; cnt < length; cnt++) {
    regData = kBtIICModeHardware | kBtIICRate100kHz | kBtIICSCL | kBtIICSDA;
    
    regData |= addr << kBtIICDB0Shift;
    regData |= ((subaddr + cnt) & 0x0FF) << kBtIICDB1Shift;
    
    // Clear the IIC Done bit.
    _writeVideoReg32(kBtIntStat, kBtIntIICDone);
    _writeVideoReg32(kBtIIC, regData);
    
    timeOut = 0;
    while (1) {
      IODelay(1000);
      regStatus = _readVideoReg32(kBtIntStat);
      if (regStatus & kBtIntIICDone) break;
      if (++timeOut == 6) break;
    }
    
    regData = _readVideoReg32(kBtIIC);
    
    if (!(regStatus & kBtStatusRACK)) {
      ret = kIOReturnError;
      break;
    }
    
    data[cnt] = (regData >> kBtIICDB1Shift) & 0x0FF;
  }
  
  return ret;
}


// DMA programing methods
void AppleBt8xx::_setupVideoDMA(void)
{
  UInt32 cnt, pos, cmd, addr, byteCountMax, byteCount, skipCount;
  UInt32 pixelSize, rowBytes, capCtl;
  
  rowBytes = _screenInfo.rowBytes;
  pixelSize = _screenInfo.pixelSize;
  
  addr = _screenInfo.baseAddress + (_top * rowBytes) + (_left * pixelSize);
  
  byteCountMax = _width * pixelSize;
  
  pos = 0;
  
  // 0x000 Sync until start of the odd field.
  cmd = kBtDMAOpcodeSync | kBtDMAResync | kBtDMARiskStatusVRE;
  OSWriteLittleInt32(_videoCmdPage, pos, cmd); pos += 4;
  OSWriteLittleInt32(_videoCmdPage, pos, 0); pos += 4;
  
  // 0x008 Jump to odd field.
  cmd = kBtDMAOpcodeJump;
  OSWriteLittleInt32(_videoCmdPage, pos, cmd); pos += 4;
  OSWriteLittleInt32(_videoCmdPage, pos, _videoOddPagePhys); pos += 4;
  
  // 0x010 Sync until start of the even field.
  cmd = kBtDMAOpcodeSync | kBtDMAResync | kBtDMARiskStatusVRO;
  OSWriteLittleInt32(_videoCmdPage, pos, cmd); pos += 4;
  OSWriteLittleInt32(_videoCmdPage, pos, 0); pos += 4;
  
  // 0x018 Jump to even field.
  cmd = kBtDMAOpcodeJump;
  OSWriteLittleInt32(_videoCmdPage, pos, cmd); pos += 4;
  OSWriteLittleInt32(_videoCmdPage, pos, _videoEvenPagePhys); pos += 4;
  
  // 0x020 Jump back to the top.
  cmd = kBtDMAOpcodeJump;
  OSWriteLittleInt32(_videoCmdPage, pos, cmd); pos += 4;
  OSWriteLittleInt32(_videoCmdPage, pos, _videoCmdPagePhys); pos += 4;
  
  pos = 0;
  
  // 0x000 Sync until FM1.
  cmd = kBtDMAOpcodeSync | kBtDMARiskStatusFM1;
  OSWriteLittleInt32(_videoOddPage, pos, cmd);
  OSWriteLittleInt32(_videoEvenPage, pos, cmd);
  OSWriteLittleInt32(_videoOddPage, pos + 4, 0);
  OSWriteLittleInt32(_videoEvenPage, pos + 4, 0);
  pos += 8;
  
  for (cnt = 0; cnt < _heigth; cnt += 2) {
    byteCount = 0;
    if ((_top + cnt) < _screenInfo.height) {
      byteCount = byteCountMax;
      if ((_left + _width) >= _screenInfo.width) {
	byteCount = (_screenInfo.width - _left) * pixelSize;
      }
    }
    skipCount = byteCountMax - byteCount;
    
    if (byteCount != 0) {
      cmd = kBtDMAOpcodeWrite | kBtDMASOL | (byteCount & kBtDMAByteCountMask);
      if (skipCount == 0) cmd |= kBtDMAEOL;
      OSWriteLittleInt32(_videoOddPage, pos, cmd);
      OSWriteLittleInt32(_videoEvenPage, pos, cmd);
      OSWriteLittleInt32(_videoOddPage, pos + 4, addr);
      OSWriteLittleInt32(_videoEvenPage, pos + 4, addr + rowBytes);
      pos += 8;
    }
    
    if (skipCount != 0) {
      cmd = kBtDMAOpcodeSkip | kBtDMAEOL | (byteCount & kBtDMAByteCountMask);
      if (byteCount == 0) cmd |= kBtDMASOL;
      OSWriteLittleInt32(_videoOddPage, pos, cmd);
      OSWriteLittleInt32(_videoEvenPage, pos, cmd);
      pos += 4;
    }
    
    addr += rowBytes * 2;
  }
  
  cmd = kBtDMAOpcodeJump;
  OSWriteLittleInt32(_videoOddPage, pos, cmd);
  OSWriteLittleInt32(_videoEvenPage, pos, cmd);
  OSWriteLittleInt32(_videoOddPage, pos + 4, _videoCmdPagePhys + 0x010);
  OSWriteLittleInt32(_videoEvenPage, pos + 4, _videoCmdPagePhys + 0x020);
  pos += 8;
  
  OSSynchronizeIO();
  
  capCtl = kBtCapCtlCapEven | kBtCapCtlCapOdd;
  if (pixelSize == 1) capCtl |= kBtCapCtlDithFrame;
  _writeVideoReg8(kBtCapCtl, capCtl);
}


// Register accessor methods
UInt8 AppleBt8xx::_readVideoReg8(UInt32 reg)
{
  return *(UInt8 *)(videoBaseAddress + reg);
}

UInt32 AppleBt8xx::_readVideoReg32(UInt32 reg)
{
  return OSReadLittleInt32(videoBaseAddress, reg);
}

void AppleBt8xx::_writeVideoReg8(UInt32 reg, UInt8 data)
{
  *(UInt8 *)(videoBaseAddress + reg) = data;
  OSSynchronizeIO();
}

void AppleBt8xx::_writeVideoReg32(UInt32 reg, UInt32 data)
{
  OSWriteLittleInt32(videoBaseAddress, reg, data);
  OSSynchronizeIO();
}

UInt8 AppleBt8xx::_readAudioReg8(UInt32 reg)
{
  return *(UInt8 *)(audioBaseAddress + reg);
}

UInt32 AppleBt8xx::_readAudioReg32(UInt32 reg)
{
  return OSReadLittleInt32(audioBaseAddress, reg);
}

void AppleBt8xx::_writeAudioReg8(UInt32 reg, UInt8 data)
{
  *(UInt8 *)(audioBaseAddress + reg) = data;
  OSSynchronizeIO();
}

void AppleBt8xx::_writeAudioReg32(UInt32 reg, UInt32 data)
{
  OSWriteLittleInt32(audioBaseAddress, reg, data);
  OSSynchronizeIO();
}


//---------------------------------------------------------------------------//

IOReturn AppleBt8xxUserClient::_extCmdNone(void)
{
  return kIOReturnSuccess;
}


IOReturn AppleBt8xxUserClient::_extCmdSetup(AppleBt8xxCmdSetupInfoPtr info)
{
  return owner->_setup();
}


IOReturn AppleBt8xxUserClient::_extCmdStart(void)
{
  return owner->_start();
}

IOReturn AppleBt8xxUserClient::_extCmdStop(void)
{
  return owner->_stop();
}


IOReturn AppleBt8xxUserClient::_extCmdGetSource(SInt32 *source)
{
  *source = owner->_currentSource;
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xxUserClient::_extCmdSetSource(SInt32 source)
{
  return owner->_setSource(source);
}


IOReturn AppleBt8xxUserClient::_extCmdSetTuner(UInt32 frequency)
{
  return owner->_setTuner(frequency);
}

IOReturn AppleBt8xxUserClient::_extCmdSetVideoControl(eAppleBt8xxControl control, UInt32 value)
{
  IOReturn ret;
  
  switch (control) {
  case kAppleBt8xxControlBrightness:
    ret = owner->_setBrightness((SInt8)(value & 0xFF));
    break;
    
  case kAppleBt8xxControlContrast:
    ret = owner->_setContrast((UInt16)(value & 0xFFFF));
    break;
    
  case kAppleBt8xxControlSaturationU :
    ret = owner->_setSaturationU((UInt16)(value & 0xFFFF));
    break;
    
  case kAppleBt8xxControlSaturationV :
    ret = owner->_setSaturationV((UInt16)(value & 0xFFFF));
    break;
    
  case kAppleBt8xxControlHue :
    ret = owner->_setHue((SInt8)(value & 0xFF));
    break;
    
  default :
    ret = kIOReturnBadArgument;
    break;
  }
  
  return ret;
}

IOReturn AppleBt8xxUserClient::_extCmdReadReg(UInt32 reg, UInt32 *data)
{
  *data = owner->_readVideoReg32(reg);
  
  return kIOReturnSuccess;
}

IOReturn AppleBt8xxUserClient::_extCmdWriteReg(UInt32 reg, UInt32 data)
{
  owner->_writeVideoReg32(reg, data);
  
  return kIOReturnSuccess;
}
